/*
 *    env_audit - tests and records everything it can about its environment
 *    This program is copyright (c) 2001-2002
 *    by Steve Grubb <linux_4ever@yahoo.com> ALL RIGHTS RESERVED
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published  by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 *    Released 1.05  11/11/02
 *    Released 1.04  10/21/02
 *    Released 1.03  10/15/02
 *    Released 1.02  07/06/01
 */

#include <stdio.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <string.h>
#include <sys/resource.h>
#include <pwd.h>
#include <grp.h>
#include <termios.h>
#include <stropts.h>
#ifdef S_ISDOOR
#include <door.h>
#endif
#ifdef _POSIX_CAP
#include <sys/capability.h>
#endif

#ifndef PATH_MAX
#define PATH_MAX 768
#endif

#ifndef SIGUNUSED
#define SIGUNUSED NSIG-1
#endif

//#define NO_SIGTEST    // <--- Uncomment if signal test annoys you 
#define VERSION "1.05"
#define LOGFILE "/tmp/env_audit.log"
extern char **environ;

union val {
  int i_val;
  long l_val;
  char c_val[10];
  struct linger linger_val;
  struct timeval timeval_val;
#ifdef SO_PEERCRED
  struct ucred peercred;
#endif
} val;

typedef union tsockaddr {
   struct sockaddr     sa;
   struct sockaddr_in  sa_in;
   struct sockaddr_in6 sa_in6;
   char   pad[128];
}xsockaddr;

static char *str_flag(const union val *, int);
static char *str_int(const union val *, int);
static char *str_linger(const union val *, int);
static char *str_timeval(const union val *, int);
#ifdef SO_PEERCRED
static char *str_cred(const union val *, int);
#endif
static void pr_limits(const char *name, int resource);
#define print_limits(name) pr_limits(#name, name)
static char *lookup_uid(int);
static char *lookup_grp(int);
static const char *lookup_family(int type);
static void output_stream(int fd);
static char *find_inode(int d, int leaked);

struct sock_opts {
  const char *opt_str;
  int opt_level;
  int opt_name;
  char *(*opt_val_str)(const union val *, int);
} sock_opts[] = {
	{"SO_BROADCAST", SOL_SOCKET, SO_BROADCAST, str_flag},
	{"SO_DEBUG", SOL_SOCKET, SO_DEBUG, str_flag},
	{"SO_DONTROUTE", SOL_SOCKET, SO_DONTROUTE, str_flag},
	{"SO_ERROR", SOL_SOCKET, SO_ERROR, str_int},
	{"SO_KEEPALIVE", SOL_SOCKET, SO_KEEPALIVE, str_flag},
	{"SO_LINGER", SOL_SOCKET, SO_LINGER, str_linger},
	{"SO_OOBINLINE", SOL_SOCKET, SO_OOBINLINE, str_flag},
	{"SO_RCVBUF", SOL_SOCKET, SO_RCVBUF, str_int},
	{"SO_SNDBUF", SOL_SOCKET, SO_SNDBUF, str_int},
	{"SO_RCVLOWAT", SOL_SOCKET, SO_RCVLOWAT, str_int},
	{"SO_SNDLOWAT", SOL_SOCKET, SO_SNDLOWAT, str_int},
	{"SO_RCVTIMEO", SOL_SOCKET, SO_RCVTIMEO, str_timeval},
	{"SO_SNDTIMEO", SOL_SOCKET, SO_SNDTIMEO, str_timeval},
	{"SO_REUSEADDR", SOL_SOCKET, SO_REUSEADDR, str_flag},
#ifdef SO_REUSEPORT
	{"SO_REUSEPORT", SOL_SOCKET, SO_REUSEPORT, str_flag},
#else
	{"SO_REUSEPORT", 0, 0, NULL},
#endif
	{"SO_TYPE", SOL_SOCKET, SO_TYPE, str_int},
#ifdef SO_USELOOPBACK
	{"SO_USELOOPBACK", SOL_SOCKET, SO_USELOOPBACK, str_flag},
#else
	{"SO_USELOOPBACK", 0, 0, NULL},
#endif
	{"IP_TOS", IPPROTO_IP, IP_TOS, str_int},
	{"IP_TTL", IPPROTO_IP, IP_TTL, str_int},
#ifdef IPV6_ADDRFORM
	{"IPV6_ADDRFORM", IPPROTO_IPV6, IPV6_ADDRFORM, str_flag},
#else
	{"IPV6_ADDRFORM", 0, 0, NULL},
#endif
#ifdef IPV6_IPV6ONLY
	{"IPV6_IPV6ONLY", IPPROTO_IPV6, IPV6_IPV6ONLY, str_flag},
#else
	{"IPV6_IPV6ONLY", 0, 0, NULL},
#endif
	{"TCP_MAXSEG", IPPROTO_TCP, TCP_MAXSEG, str_int},
	{"TCP_NODELAY", IPPROTO_TCP, TCP_NODELAY, str_flag},
#ifdef SO_PEERCRED
	{"SO_PEERCRED", SOL_SOCKET, SO_PEERCRED, str_cred},
#endif
	{NULL, 0, 0, NULL}
};

FILE *f = 0;
static char found_file[PATH_MAX+1];

int main(int argc, char *argv[])
{
	int i, fn;
	char *ptr, **pptr, cwd_buf[128];
	sigset_t sigset;

	/*
	 * Check to see if this is being run from a web server
	 */
	if (getenv("HTTP_ACCEPT"))
	{
		f = stdout;
		fprintf(f, "Content-type: text/plain\n\n");
	}
	else
	{
		unlink(LOGFILE);
		f = fopen(LOGFILE, "wt");
		if (f == NULL)
			return 1;
	}
	fn = fileno(f); 
	fprintf(f, "Environment Audit version %s\n\n", VERSION);

	/*
	 * We start of looking at the process ID stuff
	 */
	fprintf(f, "Process ID: %d\n", getpid());
	fprintf(f, "Parent Process ID: %d\n", getppid());
	fprintf(f, "User ID: %d - %s\n", getuid(), lookup_uid(getuid()));
	fprintf(f, "Group ID: %d - %s\n", getgid(), lookup_grp(getgid()));
	fprintf(f, "Effective User ID: %d - %s\n", geteuid(), 
			lookup_uid(geteuid()) );
	if ((geteuid() == 0) && (geteuid() != getuid()))
		fprintf(f, 
			"Security hole - Effective ID is root and UID isn't\n");
	fprintf(f, "Effective Group ID: %d - %s\n", getegid(), 
			lookup_grp(getegid()));
	if ((getegid() == 0) && (getegid() != getgid()))
		fprintf(f, 
		  "Security hole - Effective Group ID is root and GID isn't\n");
	if (getgroups(0,0) > 0)
	{
		gid_t *list;
		int num_groups = getgroups(0,0);

		fprintf(f, "Supplemental Groups: ");
		list = (gid_t *)malloc(sizeof(gid_t)*num_groups);
		if (list)
		{
			int op = 0, my_gid = getgid();
			getgroups(num_groups, list);
			for (i=0; i<num_groups; i++)
			{
				if (i == my_gid)
					continue;
				if (op)
					fprintf(f, ", %s", lookup_grp(list[i]));
				else
				{
					fprintf(f, " %s", lookup_grp(list[i]));
					op = 1;
				}
			}
			free(list);
		}
		fprintf(f, "\n");	
	}
	if (getuid() != 0)
	{
		if (setuid(0) == 0)
			fprintf(f, "Warning - was able to regain root!\n");
	}
	fprintf(f, "Process Group ID: %d\n", getpgrp());
	fprintf(f, "Session ID: %d\n", getsid(getpid()));
	fprintf(f, "Parent Session ID: %d\n", getsid(getppid()));
	getcwd(cwd_buf, 127);
	fprintf(f, "Current Working Dir: %s\n", cwd_buf);
	fprintf(f, "Umask: %o\n", umask(777));
	fprintf(f, "Process Priority: %d\n", getpriority(PRIO_PROCESS, 0));

	/*
	 * Now, lets take a look at the inherited signals
	 */
	sigprocmask(0, NULL, &sigset);
	for (i=1; i<=SIGUNUSED; i++)
	{
		if (sigismember(&sigset, i))
			fprintf(f, "Signal %d-%s masked\n", i, strsignal(i));
	}
#ifndef NO_SIGTEST
	/*
	 *  Make sure we aren't sending signals to init just in case it 
	 *  doesn't handle signals well...it could be impossible to boot
	 *  the system if env_audit is in the inittab and init dies when
	 *  SIGUSR1 is sent to it.
	 */
	if (getppid() > 2)
	{
		if (kill(getppid(), SIGUSR1) == 0)
			fprintf(f, 
		 "WARNING the child program can send signals to the parent.\n");
	}
#endif
	/* Dump Command Line arguments */
	if (argc > 1)
	{
		fprintf(f, "\n---\nCommand Line Arguments\n");
		for (i=1; i<argc; i++)
			fprintf(f, "Command Line Arg %d: %s\n", i, argv[i]);
	}

	/* Dump environmental variables */
	fprintf(f, "\n---\nEnvironmental Variables\n");
	pptr = environ;
	while(*pptr)
		fprintf(f, "$%s\n", *pptr++);
	
	ptr = getenv("LD_LIBRARY_PATH");
	if (ptr)
		fprintf(f, "WARNING LD_LIBRARY_PATH is defined.\n");
	ptr = getenv("LD_PRELOAD");
	if (ptr)
		fprintf(f, "WARNING LD_PRELOAD is defined.\n");
	ptr = getenv("IFS");
	if (!ptr)
		fprintf(f, "WARNING $IFS undefined\n");
	ptr = getenv("TZ");
	if (!ptr)
		fprintf(f, "WARNING $TZ undefined\n");
	ptr = getenv("HOSTALIASES");
	if (ptr)
		fprintf(f, 
		   "WARNING HOSTALIASES: %s will be used for gethostbyname.\n", 
		    ptr);

	/* Examine the process rlimits */		
	fprintf(f, "\n---\nResource Limits\n");
	fprintf(f, "Name            Current     Max\n");
	print_limits(RLIMIT_CORE);
	print_limits(RLIMIT_CPU);
	print_limits(RLIMIT_DATA);
	print_limits(RLIMIT_FSIZE);
#ifdef RLIMIT_MEMLOCK
	print_limits(RLIMIT_MEMLOCK);
#endif
#ifdef RLIMIT_NOFILE
	print_limits(RLIMIT_NOFILE);
#endif
#ifdef RLIMIT_OFILE
	print_limits(RLIMIT_OFILE);
#endif
#ifdef RLIMIT_NPROC
	print_limits(RLIMIT_NPROC);
#endif
#ifdef RLIMIT_RSS
	print_limits(RLIMIT_RSS);
#endif
	print_limits(RLIMIT_STACK);
#ifdef RLIMIT_VMEM
	print_limits(RLIMIT_VMEM);
#else
#ifdef RLIMIT_AS
	print_limits(RLIMIT_AS);
#endif
#endif
	
#ifdef _POSIX_CAP
	// This is where Posix Capabilities will go.
	// This needs research to complete...
	fprintf(f, "\n---\nPosix Capabilities\n");
	{
		char *res = 0;
		ssize_t len = 0;
		cap_t caps = cap_get_proc();
		if (caps)
			res = cap_to_text(caps, &len);

		if (res)
		{
			if ( (len == 1)||(strcmp(res, "all=") == 0) )
				fprintf(f, "Capabilities = ALL\n");
			else
				fprintf(f, "Capabilities: '%s'\n", res);
		}
		else
			fprintf(f, "Error reading capabilities\n");
		if (caps)
			cap_free(caps);
	}
#endif
	fflush(f);

	/* Examine the file descriptors */
	for (i=0; i< 1024; i++)
	{
		struct stat st;
		int val, am, std = 0, leaked = 0;

		if ((i == fn)&&(i != fileno(stdout)))
			continue;
		if (fstat(i, &st) == -1)
			continue;
		fprintf(f, "\n---\n");
		fprintf(f, "Open file descriptor: %d\n", i);
		fprintf(f, "User ID of File Owner: %s\n", 
				lookup_uid(st.st_uid));
		fprintf(f, "Group ID of File Owner: %s\n", 
				lookup_grp(st.st_gid));
	
		if (fileno(stdin) == i)
		{
			fprintf(f, "Descriptor is stdin.\n");
			std = 1;
		}
		else if (fileno(stdout) == i)
		{
			fprintf(f, "Descriptor is stdout.\n");
			std = 1;
		}
		else if (fileno(stderr) == i)
		{
			fprintf(f, "Descriptor is stderr.\n");
			std = 1;
		}
		else
		{
			fprintf(f, 
			    "WARNING - Descriptor is leaked from parent.\n");
			leaked = 1;
		}
		if (std)
		{
			int fore;
			struct termios term;

			fore = tcgetattr(i, &term);
			if (fore != -1)
			{
				fprintf(f, 
				"WARNING - process has controlling terminal\n");
			}
			else
			{
				fprintf(f, "No controlling terminal\n");
			}
		}
		fprintf(f, "File type: ");
		if (S_ISREG(st.st_mode))
		{
			char *filename;
			fprintf(f, "regular file, inode: %ld, device: %d\n", 
					st.st_ino, (int)st.st_dev);
			filename = find_inode(i, leaked);
			if (filename)
			{
				if (stat(filename, &st) == 0)
				{
					fprintf(f, 
					    "File's actual permissions: %o\n",
						st.st_mode & 
						(S_ISUID | S_ISGID | S_ISVTX |
						 S_IRWXU | S_IRWXG | S_IRWXO));
				}
				else
					fprintf(f, "Cannot stat actual file\n");
			}
		}
		else if (S_ISDIR(st.st_mode))
			fprintf(f, "directory\n");
		else if (S_ISCHR(st.st_mode))
		{
			char *tptr;

			tptr = ttyname(i);
			if (tptr)
			{
				fprintf(f, "character device\n");
				fprintf(f, 
				"The tty connected to this descriptor is: %s\n",
				tptr);
			}
			else
			{
				fprintf(f, "character device, inode: %ld\n", 
						st.st_ino);
				find_inode(i, leaked); 
			}
		}
		else if (S_ISBLK(st.st_mode))
		{
			fprintf(f, "block device, inode: %ld\n", 
					st.st_ino);
			find_inode(i, leaked); 
		}
		else if (S_ISFIFO(st.st_mode))
		{
			if (st.st_nlink == 0)
				fprintf(f, "pipe\n");
			else
			{
				fprintf(f, "fifo, inode: %ld, device: %d\n", 
					st.st_ino, (int)st.st_dev);
				find_inode(i, leaked);
			}
		}
		else if (S_ISLNK(st.st_mode))
			fprintf(f, "symbolic link\n");
		else if (S_ISSOCK(st.st_mode))
		{
			xsockaddr localaddr, peeraddr;
			int llen, plen, port;
			struct sock_opts *sptr;
			char addr[INET6_ADDRSTRLEN];

			llen = sizeof(localaddr);
			plen = sizeof(peeraddr);
			fprintf(f, "socket\n");
			getsockname(i, &localaddr.sa, &llen);
			getpeername(i, &peeraddr.sa, &plen);
			
			if (localaddr.sa.sa_family == AF_INET)
			{
				inet_ntop(AF_INET, &localaddr.sa_in.sin_addr,
					addr, sizeof(addr));
				port = ntohs(localaddr.sa_in.sin_port);
			}
			else if (localaddr.sa.sa_family == AF_INET6)
			{
				inet_ntop(AF_INET6, &localaddr.sa_in6.sin6_addr,
					addr, sizeof(addr));
				port = ntohs(localaddr.sa_in6.sin6_port);
			}
			else if (localaddr.sa.sa_family == AF_UNIX)
			{
				port = 0;
				strcpy(addr, "Unix Domain");
			}
			else
			{
				strcpy(addr, "UNKNOWN");
				port = 0;
			}
			fprintf(f,"Address Family: %s\n", 
				lookup_family(localaddr.sa.sa_family));
			fprintf(f,"Local address: %s\n", addr);
			if (port)
			{
				if (port < 1024)
				{	// Look up service name
					struct servent *serv = 
						getservbyport(htons(port), 
							NULL);
					if (serv)
					    fprintf(f,"Local Port: %d, %s\n", 
						port, serv->s_name);
					else
					    fprintf(f,"Local Port: %d\n", port);
					endservent();
					fprintf(f,
						  "NOTICE - connected to a "
						  "privileged port\n");
				}
				else
					fprintf(f,"Local Port: %d\n", port);
			}
			if (peeraddr.sa.sa_family == AF_INET)
			{
				inet_ntop(AF_INET, &peeraddr.sa_in.sin_addr,
					addr, sizeof(addr));
				port = ntohs(peeraddr.sa_in.sin_port);
			}
			else if (peeraddr.sa.sa_family == AF_INET6)
			{
				inet_ntop(AF_INET6, &peeraddr.sa_in6.sin6_addr,
					addr, sizeof(addr));
				port = ntohs(peeraddr.sa_in6.sin6_port);
			}
			else if (peeraddr.sa.sa_family == AF_UNIX)
			{
				port = 0;
				strcpy(addr, "Unix Domain");
			}
			else
			{
				strcpy(addr, "UNKNOWN");
				port = 0;
			}
			fprintf(f,"Peer address: %s\n", addr);
			if (port)
			{
				if (port < 1024)
				{	// Look up service name
					struct servent *serv = 
						getservbyport(htons(port), 
							NULL);
					if (serv)
					    fprintf(f,"Peer Port: %d, %s\n", 
						port, serv->s_name);
					else
					    fprintf(f,"Peer Port: %d\n", port);
					endservent();
					fprintf(f,
						  "NOTICE - connected to a "
						  "privileged port\n");
				}
				else
					fprintf(f,"Peer Port: %d\n", port);
			}
			fprintf(f, "Socket options:\n");
			for (sptr = sock_opts; sptr->opt_str; sptr++)
			{
#ifdef IPV6_ADDRFORM
				if ((sptr->opt_name == IPV6_ADDRFORM) &&
					(localaddr.sa.sa_family != AF_INET6))
					continue;
#endif
#ifdef IPV6_IPV6ONLY
				if ((sptr->opt_name == IPV6_IPV6ONLY) &&
					(localaddr.sa.sa_family != AF_INET6))
					continue;
#endif
				fprintf(f, "\t%s: ", sptr->opt_str);
				if (sptr->opt_val_str == NULL)
					fprintf(f, "undefined\n");
				else
				{
					union val value;
					int len = sizeof(val);
					if (getsockopt(i, sptr->opt_level, 
							sptr->opt_name, 
							&value, &len) 
							== -1)
						fprintf(f, 
							"getsockopt error.\n");
					else
						fprintf(f, "%s\n", 
							(*sptr->opt_val_str)
							(&value, len));
				}
			}
		}
#ifdef S_ISDOOR
		else if (S_ISDOOR(st.st_mode))
		{
			struct door_info info;

			fprintf(f, "door ipc\n");
			if (door_info(i, &info) == 0)
			{
				fprintf(f, "door server pid: %d\n", 
					info.di_target);
				fprintf(f, "door attributes: ");
				if (info.di_attributes & DOOR_LOCAL)
					fprintf(f, "DOOR_LOCAL ");
				if (info.di_attributes & DOOR_UNREF)
					fprintf(f, "DOOR_UNREF ");
				if (info.di_attributes & DOOR_UNREF_MULTI)
					fprintf(f, "DOOR_UNREF_MULTI ");
				if (info.di_attributes & DOOR_IS_UNREF)
					fprintf(f, "DOOR_IS_UNREF ");
				if (info.di_attributes & DOOR_REVOKED)
					fprintf(f, "DOOR_REVOKED ");
				if (info.di_attributes & DOOR_PRIVATE)
					fprintf(f, "DOOR_PRIVATE ");
				fprintf(f, "\n");
			}
		}
#endif
		else 
			fprintf(f, "unknown\n");
		val = fcntl(i, F_GETOWN, 0);
		if (val > 0)
			fprintf(f,
				"Process ID that receives SIGIO & SIGURG :%d\n",
				val);
		val = fcntl(i, F_GETFL, 0);
		am = val & O_ACCMODE;
		fprintf(f, "File descriptor is ");
		if (am == O_RDONLY)
			fprintf(f, "read only");
		else if (am == O_WRONLY)
			fprintf(f, "write only");
		else if (am == O_RDWR)
			fprintf(f, "read and write");
		else
			fprintf(f, "unknown access mode");
		if (val & O_APPEND)
			fprintf(f, ", append");
		if (val & O_NONBLOCK)
			fprintf(f, ", non-blocking");
#ifdef O_SYNC
		if (val & O_SYNC)
			fprintf(f, ", synchronous writes");
#endif
		fprintf(f, "\n");
		// Note: if the effective id is 0 or the same as the
		// file's owner, we can change this file's mode. We
		// don't actually do it since that could be destructive.
		// We just go through the normal chacks and output the results
		// This could be really bad if root owns the file
		// since it could be made setuid.
		if (geteuid() == 0 || geteuid() == st.st_uid || 
				getegid() == 0 || getegid() == st.st_gid)
			fprintf(f, "WARNING - application can execute a "
				"fchmod on this descriptor\n");
		if (isastream(i))
			output_stream(i);
		fflush(f);
	}
	fprintf(f, "\n---\n");
	fprintf(f, "Audit Complete\n\n");
	fclose(f);
	return 0;
}

static char strres[128];
static char *str_flag(const union val *ptr, int len)
{
	if (len != sizeof(int))
		snprintf(strres, 128, "size (%d) not sizeof(int)", len);
	else
		snprintf(strres, 128, "%s", (ptr->i_val == 0) ? "off" : "on");
	return strres;
}

static char *str_int(const union val *ptr, int len)
{
	if (len != sizeof(int))
		snprintf(strres, 128, "size (%d) not sizeof(int)", len);
	else
		snprintf(strres, 128, "%d", ptr->i_val);
	return strres;
}

static char *str_linger(const union val *ptr, int len)
{
	if (ptr->linger_val.l_onoff == 0)
		snprintf(strres, 128, "off");
	else
		snprintf(strres, 128, "%d seconds", ptr->linger_val.l_linger);
	return strres;
}

static char *str_timeval(const union val *ptr, int len)
{
	snprintf(strres, 128, "%d seconds and %d microseconds", 
			(int)ptr->timeval_val.tv_sec, 
			(int)ptr->timeval_val.tv_usec);
	return strres;
}

#ifdef SO_PEERCRED
static char *str_cred(const union val *ptr, int len)
{
	snprintf(strres, 128, "peer uid %d, peer gid %d", 
			(int)ptr->peercred.uid, 
			(int)ptr->peercred.gid);
	return strres;
}
#endif

static void pr_limits(const char *name, int resource)
{
	struct rlimit limit;

	if (getrlimit(resource, &limit) < 0)
	{
		fprintf(f, "getrlimit error for %s\n", name);
		return;
	}
	fprintf(f, "%-14s  ", name);
	if (limit.rlim_cur == RLIM_INFINITY)
		fprintf(f, "(infinity)  ");
	else
		fprintf(f, "%10ld  ", limit.rlim_cur);
	if (limit.rlim_max == RLIM_INFINITY)
		fprintf(f, "(infinity)\n");
	else
		fprintf(f, "%10ld\n", limit.rlim_max);
}

static char unknown[] = "UNKNOWN", id[33];
static char *lookup_uid(int uid)
{
	struct passwd *pass;

	pass = getpwuid(uid);
	if (pass == NULL)
		return unknown;
	strncpy(id, pass->pw_name, 32);
	id[32] = 0;
	endpwent();
	return id;
}

static char *lookup_grp(int gid)
{
	struct group *gp;

	gp = getgrgid(gid);
	if (gp == NULL)
		return unknown;
	strncpy(id, gp->gr_name, 32);
	id[32] = 0;
	endgrent();
	return id;
}

static const char *lookup_family(int type)
{
	const char *ret_val;
	static char buf[64];
	
	switch (type)
	{
	case AF_INET:
		ret_val = "AF_INET";
		break;
	case AF_INET6:
		ret_val = "AF_INET6";
		break;
	case AF_UNIX:
		ret_val = "AF_UNIX";
		break;
	default:
		sprintf(buf, "UNKNOWN #%d\n", type);
		ret_val = buf;
		break;
	}
	return ret_val;
}

static void output_stream(int fd)
{
	struct str_list list;
	int idx, nmods;

	if ( (nmods = ioctl(fd, I_LIST, (void *)0)) >= 0)
	{
		fprintf(f, "The descriptor is also a stream\n");
		fprintf(f, " num of stream modules: %d\n", nmods);

		/* allocate space for modules */
		list.sl_modlist = calloc(nmods, sizeof(struct str_mlist));
		if (list.sl_modlist)
		{
			list.sl_nmods = nmods;

			/* fetch the module names */
			if (ioctl(fd, I_LIST, &list) >= 0)
			{
				for (idx=0; idx<=nmods; idx++)
					fprintf( f, 
					   "  %s: %s\n", 
					   (idx == nmods) ? "driver" : "module",
					   list.sl_modlist++
					);
			}
		}
	}
}

static char *fullpath;
static char *find_inode(int d, int leaked)
{
	fullpath = malloc(PATH_MAX+1);
	if (fullpath == 0)
		return NULL;
	memset(found_file, 0, PATH_MAX);

	snprintf(fullpath, PATH_MAX, "/proc/%d/fd/%d", getpid(), d);
	if (readlink(fullpath, found_file, PATH_MAX) >= 0)
		found_file[PATH_MAX] = 0;
        free(fullpath);
										        if (found_file[0] == 0)
        {
                fprintf(f, "Could not find exact descriptor.\n");
                return NULL;
        }
        else
	{
		int offset = 0;
		if (found_file[1] == '/')
			offset = 1;
                if (leaked == 1)
                        fprintf(f, "The leaked descriptor is: %s\n",
                                   found_file+offset);
                else if (leaked == 0)
                        fprintf(f, "The descriptor is: %s\n",
                                   found_file+offset);
                return found_file+offset;
        }
}
												
